classdef hallBasis
    %HALLBASIS - A class for Philipp Hall bases of Lie brackets. 
    %
    %     obj = hallBasis(vecFieldIDs,maxDegree,optOnlyDiffVecFields)
    %     obj = hallBasis(vecFieldIDs,maxDegree,optOnlyDiffVecFields,optOnlyNonZeroVecFields,vecFields)
    %
    %Inputs
    %- vecFieldsIDs                 A vector containing the identifiers
    %                               (integer values) of the vector fields
    %                               for which the P. Hall basis shall be
    %                               computed.
    %- maxDegree                    The maximal degree of the brackets in
    %                               the P. Hall basis. 
    %- optOnlyDifferentVecFields    Set 'true' when only brackets with 
    %                               distinct vector fields shall be
    %                               computed and set 'false' if all
    %                               brackets shall be computed. For
    %                               example, in the first case a bracket of
    %                               the form [1,[1,4]] would not be
    %                               computed whereas it would in the second
    %                               case.
    %- optOnlyNonZeroVecFields      Set 'true' if only those brackets shall
    %                               be computed that do not evaluate to
    %                               zero for the given set of vector fields
    %                               'symVecFields'. If set to 'false',
    %                               this gives the same as without the last
    %                               two arguments.
    %- symVecFields                 An array of objects of class
    %                               'vectorField' containing the vector
    %                               fields for which the brackets shall be
    %                               evaluated.
    %
    %== PROPERTIES ==
    %
    %Property name       Data type     Description
    %----------------------------------------------------------------------
    %- brackets          bracket       Contains all brackets that form the 
    %                                  Hall basis. 
    %- startIndexDegree  pos. integer  Vector of starting indices for brackets 
    %                                  of different degree, i.e., all brackets 
    %                                  with index from startIndexDegree(i) to
    %                                  startIndexDegree(i+1)-1 have degree i.
    %                                  Use the method 'getBracket(i)' to return 
    %                                  all these brackets.
    %- maxDegree         pos. integer  Maximal degree of all brackets in
    %                                  the basis.
    %- vecFieldIDs       pos. integer  IDs of the vector fields from which
    %                                  the basis was formed.
    %
    %== METHODS ==
    %
    %- printBasis        Display all brackets in the basis.
    %- getBrackets       Get all brackets of a given degree.
    %- getID             Search for the ID of a given bracket.
    %
    %S. Michalowsky, August 2017
    
    properties (SetAccess = private, GetAccess = public)
        brackets            % Array of object of class 'bracket', contains 
                            % all brackets in the basis.
        startIndexDegree    % Vector of starting indices for brackets of 
                            % different degree, i.e., all brackets with
                            % index from startIndexDegree(i) to
                            % startIndexDegree(i+1)-1 have degree i. Use
                            % the method 'getBracket(i)' to return all
                            % these brackets.
        maxDegree           % Pos. integers that contains the maximal degree 
                            % of all brackets in the basis.
        vecFieldIDs         % IDs of the vector fields in the Hall basis
    end
    
    methods
        % CONSTRUCTOR
        function obj = hallBasis(vecFieldIDs,maxDegree,optOnlyDiffVecFields,varargin)
            %Compute Hall basis for the given vector fields.
            
            
            % get possible additional arguments
            if nargin == 5
                optOnlyNonZeroVecFields = varargin{1};
                vecFields = varargin{2};
            else
                optOnlyNonZeroVecFields = false;
            end
            
            % maximal degree
            obj.maxDegree = maxDegree;
            
            % sort vector fields
            vecFieldIDs = sort(vecFieldIDs);
            obj.vecFieldIDs = vecFieldIDs;
            nVecFields = length(vecFieldIDs);
            
            % first elements: all vector fields
            disp(['Computing brackets of degree 1/',num2str(maxDegree),'...'])
            theStartIndexDegree = 1;
            obj.brackets = bracket(vecFieldIDs(1),[],1); %vecFieldIDs(1));
            for iVecField = 2:nVecFields
                obj.brackets(iVecField) = bracket(vecFieldIDs(iVecField),[],length(obj.brackets)+1); %vecFieldIDs(iVecField));
            end
            disp('Finished.')
            
            % brackets of degree 2
            disp(['Computing brackets of degree 2/',num2str(maxDegree),'...'])
            theStartIndexDegree(2) = length(obj.brackets) + 1;
            if maxDegree > 1
                for iVecField1 = 1:nVecFields
                    for iVecField2 = iVecField1+1:nVecFields
                        if optOnlyNonZeroVecFields
                            theBracket = bracket( obj.brackets(iVecField1), obj.brackets(iVecField2));
                            % check whether bracket is zero
                            isZeroBracket = theBracket.isZero(vecFields);
                            if ~isZeroBracket
                                obj.brackets(end+1) = bracket( obj.brackets(iVecField1), obj.brackets(iVecField2), length(obj.brackets)+1);
                            end
                        else
                            obj.brackets(end+1) = bracket( obj.brackets(iVecField1), obj.brackets(iVecField2), length(obj.brackets)+1);
                        end
                    end
                end
            end
            disp('Finished.')
            
            % brackets of degree >= 3
            for iDegree = 3:maxDegree
                disp(['Computing brackets of degree ',num2str(iDegree),'/',num2str(maxDegree),'...'])
                theStartIndexDegree(iDegree) = length(obj.brackets) + 1;
                % iterate over all brackets with degree <= iDegree/2
                for iBracket = 1:theStartIndexDegree(floor(iDegree/2)+1)-1
                    bracket1 = obj.brackets(iBracket);                  % left factor of bracket 
                    degBracket1 = obj.brackets(iBracket).degree;        % degree of left factor
                    degBracket2 = iDegree - degBracket1;                % required degree of right factor
                    % iterate over all brackets with the required degree
                    for iBracket2 = theStartIndexDegree(degBracket2):theStartIndexDegree(degBracket2+1)-1
                        bracket2 = obj.brackets(iBracket2);
                        % check whether condition for P. Hall bracket is
                        % fulfilled
                        if ( bracket1.ID < bracket2.ID ) && ( bracket2.bracket1.ID < bracket1.ID ) 
                            if optOnlyDiffVecFields
                                % count number of equal vector fields 
                                % in bracket1 and bracket2
                                nEqualVecFields = 0;
                                for iVecField = 1:length(bracket1.vecFieldIDs)
                                    nEqualVecFields = nEqualVecFields + sum( bracket2.vecFieldIDs == bracket1.vecFieldIDs(iVecField) );
                                end
                                if nEqualVecFields == 0
                                    % all vector fields are different
                                    if optOnlyNonZeroVecFields
                                        theBracket = bracket(bracket1,bracket2);
                                        if ~theBracket.isZero(vecFields)
                                            obj.brackets(end+1) = bracket(bracket1,bracket2,length(obj.brackets)+1);
                                        end
                                    else
                                        obj.brackets(end+1) = bracket(bracket1,bracket2,length(obj.brackets)+1);
                                    end
                                end
                            else
                                if optOnlyNonZeroVecFields
                                    theBracket = bracket(bracket1,bracket2);
                                    if ~theBracket.isZero(vecFields)
                                        obj.brackets(end+1) = bracket(bracket1,bracket2,length(obj.brackets)+1);
                                    end
                                else
                                    obj.brackets(end+1) = bracket(bracket1,bracket2,length(obj.brackets)+1);
                                end
                            end
                        end
                    end
                end
                disp('Finished.')
            end
            
            % save start indices for brackets of different degrees
            obj.startIndexDegree = theStartIndexDegree;
        end
        
        % PRINT P. HALL BASIS
        function out = printBasis(obj)
            %PRINTBASIS - Display a representation of all brackets in the
            %basis. No input arguments.
            
            % Iterate over all degrees starting with one
            curDegree = 1;
            strBasis = [];
            for iBracket = 1:length(obj.brackets)
                if obj.brackets(iBracket).degree == curDegree
                    strBasis = [strBasis, obj.brackets(iBracket).toString, ', '];
                else
                    % display brackets with degree = curDegree
                    disp(['Degree ', num2str(curDegree)])
                    disp(strBasis(1:end-2))
                    fprintf('\n')
                    % increase degree
                    curDegree = obj.brackets(iBracket).degree;
                    strBasis = [];
                    strBasis = [strBasis, obj.brackets(iBracket).toString, ', '];
                end
            end
            disp(['Degree ', num2str(curDegree)])
            disp(strBasis(1:end-2))
        end
        
        % GET BRACKETS OF A CERTAIN DEGREE 
        function out = getBrackets(obj,degree)
            %GETBRACKETS - Returns all brackets of a certain degree. 
            %
            %     hallBasis.getBrackets(degree)
            
            if degree < obj.maxDegree
                out = obj.brackets( obj.startIndexDegree(degree):obj.startIndexDegree(degree+1)-1 );
            else
                if degree == obj.maxDegree
                    out = obj.brackets( obj.startIndexDegree(degree):end );
                else
                    disp('Requested degree larger than max. degree. Please recompute Hall basis.')
                end
            end
        end
        
        % GET ID OF A BRACKET
        function out = getID(obj,theBracket)
            %GETID - Returns the ID of the bracket 'theBracket'.
            %
            %     hallBasis.getID(theBracket)
            
            % get all brackets with equal degree as 'theBracket'
            allBrackets = obj.getBrackets(theBracket.degree);
            
            % go through all brackets of this degree and compare
            theID = NaN;
            for iBracket = 1:length(allBrackets)
                if allBrackets(iBracket).isEqual(theBracket)
                    theID = iBracket;
                    break;
                end
            end
            if isnan(theID)
                out = NaN;
                disp(['The requested bracket ', theBracket.toString,' could not be found in the basis.'])
            else
                out = allBrackets(iBracket).ID;
            end
        end
    end
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
end

